;;;; streams-primitive.scm
;;;; Kon Lovett, Apr '09

; Copyright (C) 2007 by Philip L. Bewig of Saint Louis, Missouri, USA.  All rights
; reserved.  Permission is hereby granted, free of charge, to any person obtaining a copy of
; this software and associated documentation files (the "Software"), to deal in the Software
; without restriction, including without limitation the rights to use, copy, modify, merge,
; publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to
; whom the Software is furnished to do so, subject to the following conditions: The above
; copyright notice and this permission notice shall be included in all copies or substantial
; portions of the Software.  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
; FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
; CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
; THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(module streams-primitive

  (;export
    ;; SRFI 41 primitive
    stream?
    stream-null stream-null?
    (stream-cons $$make-stream-pair) ;($$stream-eager $$stream-lazy $$stream-delay)
    stream-pair? stream-car stream-cdr
    stream-lambda ;($$stream-lazy)
    ;; Extras
    stream-occupied?
    ;; Common errors
    check-stream error-stream
    check-stream-occupied error-stream-occupied
    ;; WTF
    ($$stream-lazy $$make-stream-lazy)
    ($$stream-eager $$make-stream-eager)
    $$stream-delay ;($$stream-lazy $$stream-eager)
    $$make-stream-lazy
    $$make-stream-eager
    $$make-stream-pair)

  (import scheme chicken
    (only type-checks define-check+error-type)
    (only type-errors define-error-type)
    record-variants)

  (require-library type-checks type-errors record-variants)

  (include "chicken-primitive-object-inlines")
  (include "streams-inlines")

;;;

(define-record-type-variant stream (unsafe unchecked inline)
  (%make-stream prom)
  $stream?  ;ignore since %stream? conflicts with predefined inline
  (prom %stream-promise %stream-promise-set!) )

(define-check+error-type stream %stream?)

(define (make-stream-box tag obj) (%cons tag obj))
(define (stream-box-tag box) (%car box))
(define (stream-box-value box) (%cdr box))
(define (stream-box-tag-set! box tag) (%set-car!/immediate box tag))
(define (stream-box-value-set! box val) (%set-cdr! box val))

(define ($$make-stream-lazy thunk) (%make-stream (make-stream-box 'lazy thunk)))
(define ($$make-stream-eager obj) (%make-stream (make-stream-box 'eager obj)))

(define (*stream-null? stream) (eq? (stream-force stream) (stream-force stream-null)))

;;;

(define-syntax $$stream-lazy
  (syntax-rules ()
    ((_ ?expr) ($$make-stream-lazy (lambda () ?expr)) ) ) )

(define-syntax $$stream-eager
  (syntax-rules ()
    ((_ ?expr) ($$make-stream-eager ?expr) ) ) )

(define-syntax $$stream-delay
  (syntax-rules ()
    ((_ ?expr) ($$stream-lazy ($$stream-eager ?expr)) ) ) )

(define (stream-force promise)
  (let ((content (%stream-promise promise)))
    (case (stream-box-tag content)
      ((eager)
        (stream-box-value content) )
      ((lazy)
        (let* ((promise* ((stream-box-value content)))
               (content  (%stream-promise promise)))
          (unless (eq? 'eager (stream-box-tag content))
            (let ((prom (%stream-promise promise*)))
              (stream-box-tag-set! content (stream-box-tag prom))
              (stream-box-value-set! content (stream-box-value prom)) )
            (%stream-promise-set! promise* content) )
         (stream-force promise) ) ) ) ) )

(define (stream? obj) (%stream? obj))

(define stream-null ($$stream-delay (%cons 'stream 'null)))

(define (stream-null? obj) (and (%stream? obj) (*stream-null? obj)))
(define (stream-occupied? obj) (and (%stream? obj) (not (*stream-null? obj))))

(define-check+error-type stream-occupied)

(define-syntax stream-lambda
  (syntax-rules ()
    ((_ FORMALS BODY0 BODY1 ...)
     (lambda FORMALS ($$stream-lazy (let () BODY0 BODY1 ...))) ) ) )

;;

(define-record-type-variant stream-pair (unsafe unchecked inline)
  (%make-stream-pair car cdr)
  %stream-pair?
  (car %stream-car)
  (cdr %stream-cdr) )

(define ($$make-stream-pair car cdr) (%make-stream-pair car cdr))

(define-error-type stream-pair)

(define (check-stream-pair loc obj)
  (cond
    ((not (%stream? obj)) (error-stream loc obj 'stream) )
    ((*stream-null? obj)  (error-stream-occupied loc obj 'stream) )
    (else
      (let ((val (stream-force obj)))
        (if (%stream-pair? val) val
          (error-stream-pair loc val 'stream)) ) ) ) )

(define-syntax stream-cons
  (syntax-rules ()
    ((_ ?expr ?strm)
     ($$stream-eager ($$make-stream-pair ($$stream-delay ?expr) ($$stream-lazy ?strm))) ) ) )

(define (stream-pair? obj) (and (%stream? obj) (%stream-pair? (stream-force obj))))

(define (stream-car strm)
  (stream-force (%stream-car (check-stream-pair 'stream-car strm))) )

(define (stream-cdr strm)
  (%stream-cdr (check-stream-pair 'stream-cdr strm)) )

) ;module streams-primitive
